import os
from pathlib import Path, PosixPath

import py_capio_cl


def test_serialize_parse_py_capio_cl_v1(tmp_path):
    path = tmp_path / "config.json"
    workflow_name = "demo"
    file1 = PosixPath("file1.txt")
    file2 = PosixPath("file2.txt")
    file3 = PosixPath("my_command_history.txt")
    file4 = PosixPath("/tmp")

    producer = "_first"
    consumer = "_last"
    intermediate = "_middle"

    engine = py_capio_cl.Engine()
    engine.setWorkflowName(workflow_name)

    # producer/consumer graph
    engine.addProducer(file1, producer)
    engine.addConsumer(file1, intermediate)
    engine.addProducer(file2, intermediate)
    engine.addConsumer(file2, consumer)
    engine.addConsumer(file1, consumer)

    # File 1: in-memory, ON_CLOSE, UPDATE, permanent
    engine.setStoreFileInMemory(file1)
    engine.setCommitRule(file1, py_capio_cl.commit_rules.ON_CLOSE)
    engine.setCommittedCloseNumber(file1, 3)
    engine.setFireRule(file1, py_capio_cl.fire_rules.UPDATE)
    engine.setPermanent(file1, True)

    # File 2: on filesystem, ON_TERMINATION, NO_UPDATE
    engine.setStoreFileInFileSystem(file2)
    engine.setCommitRule(file2, py_capio_cl.commit_rules.ON_TERMINATION)
    engine.setFireRule(file1, py_capio_cl.fire_rules.NO_UPDATE)

    # File 3: excluded, multiple producers
    for name in [producer, consumer, intermediate]:
        engine.addProducer(file3, name)
    engine.setExclude(file3, True)

    # File 4: directory with 10 files, ON_N_FILES commit rule
    engine.setCommitRule(file4, py_capio_cl.commit_rules.ON_N_FILES)
    engine.setFireRule(file4, py_capio_cl.fire_rules.NO_UPDATE)
    engine.setDirectoryFileCount(file4, 10)
    engine.addProducer(file4, intermediate)

    engine.print()

    # Serialize
    py_capio_cl.serialize(engine, path)

    # Parse back
    new_engine = py_capio_cl.Parser.parse(path)

    assert new_engine.getWorkflowName() == workflow_name

    # Parse with memory flag
    new_engine1 = py_capio_cl.Parser.parse(path, store_only_in_memory=True)
    assert len(new_engine1.getFileToStoreInMemory()) == engine.size()

    # cleanup
    os.remove(path)


def test_parser_resolve_absolute():
    json_path = "/tmp/capio_cl_jsons/V1.0/test0.json"

    engine = py_capio_cl.Parser.parse(str(json_path), "/tmp")
    assert engine.getWorkflowName() == "test"
    for f in ["/tmp/file", "/tmp/file1", "/tmp/file2", "/tmp/file3"]:
        assert engine.contains(f)


def test_parser_exception():
    json_dir = Path("/tmp/capio_cl_jsons")
    test_filenames = [
        "",
        "ANonExistingFile",
        *(str(json_dir / f"V1_test{i}.json") for i in range(1, 24)),
    ]

    for test_file in test_filenames:
        caught = False
        try:
            wf_name, engine = py_capio_cl.Parser.parse(test_file)
        except Exception as e:
            caught = True
            typename = type(e).__name__
            assert typename == "ParserException"
            assert len(str(e)) > 0
        assert caught



import json
import socket
from functools import wraps

from py_capio_cl import CAPIO_CL_DEFAULT_WF_NAME, DEFAULT_MCAST_GROUP
from py_capio_cl import CapioCLEntry


def CapioCLRule(path: str,
                workflow_name: str = CAPIO_CL_DEFAULT_WF_NAME,
                committed: str | None = None,
                fire: str | None = None,
                close_count: int | None = None,
                directory_n_file_expected: int | None = None,
                is_directory: bool | None = None,
                is_permanent: bool | None = None,
                is_excluded: bool | None = None,
                producers: list[str] | None = None,
                consumers: list[str] | None = None,
                file_dependencies: list[str] | None = None,
                multicast_group: tuple[str, int] = DEFAULT_MCAST_GROUP
                ):
    if not path:
        raise RuntimeError("ERROR: cannot specify a CAPIO-CL rule without setting a path!")

    rule = CapioCLEntry()
    if committed:
        rule.commit_rule = committed
    if fire:
        rule.fire_rule = fire

    if close_count:
        rule.commit_on_close_count = close_count

    if directory_n_file_expected:
        rule.directory_children_count = directory_n_file_expected

    if is_directory:
        rule.is_file = not is_directory

    if is_permanent:
        rule.permanent = is_permanent

    if is_excluded:
        rule.excluded = is_excluded

    if producers:
        rule.producers = producers

    if consumers:
        rule.consumers = consumers

    if file_dependencies:
        rule.file_dependencies = file_dependencies

    request_body = dict()
    request_body["path"] = path
    request_body["workflow_name"] = workflow_name
    request_body["CapioClEntry"] = json.loads(rule.to_json())

    message = json.dumps(request_body).encode('utf-8')
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)

    ttl = 1  # no broadcast outside current subnet
    sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, ttl)

    try:
        sock.sendto(message, multicast_group)
        print(f"[[py_capio_cl]] Sent bcast message for path: {path}")
    finally:
        sock.close()

    def _capiocl_rule(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

    return _capiocl_rule



import socket
import time
from pathlib import PosixPath

import py_capio_cl


def test_instantiation():
    engine = py_capio_cl.Engine()
    assert engine.size() == 0
    engine.print()


def test_add_file_default():
    engine = py_capio_cl.Engine()
    assert engine.size() == 0
    engine.newFile("test.dat")
    assert engine.size() == 1
    assert engine.getCommitRule("test.dat") == py_capio_cl.commit_rules.ON_TERMINATION
    assert engine.getFireRule("test.dat") == py_capio_cl.fire_rules.UPDATE
    assert engine.getConsumers("test.dat") == []
    assert engine.getProducers("test.dat") == []
    assert not engine.isPermanent("test.dat")
    assert not engine.isExcluded("test.dat")
    assert engine.isFile("test.dat")
    assert not engine.isDirectory("test.dat")
    assert engine.getDirectoryFileCount("test.dat") == 0
    assert not engine.isStoredInMemory("test.dat")


def test_add_file_default_glob():
    engine = py_capio_cl.Engine()
    engine.newFile("test.*")
    assert engine.size() == 1
    assert engine.getCommitRule("test.dat") == py_capio_cl.commit_rules.ON_TERMINATION
    assert engine.getFireRule("test.dat") == py_capio_cl.fire_rules.UPDATE
    assert engine.getConsumers("test.dat") == []
    assert engine.getProducers("test.dat") == []
    assert not engine.isPermanent("test.dat")
    assert not engine.isExcluded("test.dat")
    assert engine.isFile("test.dat")
    assert not engine.isDirectory("test.dat")
    assert engine.getDirectoryFileCount("test.dat") == 0
    assert not engine.isStoredInMemory("test.dat")


def test_add_file_default_glob_question():
    engine = py_capio_cl.Engine()
    engine.newFile("test.?")
    assert engine.size() == 1
    assert engine.getCommitRule("test.1") == py_capio_cl.commit_rules.ON_TERMINATION
    assert engine.getFireRule("test.1") == py_capio_cl.fire_rules.UPDATE
    assert engine.getConsumers("test.1") == []
    assert engine.getProducers("test.1") == []
    assert not engine.isPermanent("test.1")
    assert not engine.isExcluded("test.1")
    assert engine.isFile("test.1")
    assert not engine.isDirectory("test.1")
    assert engine.getDirectoryFileCount("test.1") == 0
    assert not engine.isStoredInMemory("test.1")


def test_add_file_manually():
    engine = py_capio_cl.Engine()
    path = "test.dat"
    engine.add(path, [], [], py_capio_cl.commit_rules.ON_TERMINATION,
               py_capio_cl.fire_rules.UPDATE, False, False, [])
    assert engine.size() == 1
    assert engine.getCommitRule("test.dat") == py_capio_cl.commit_rules.ON_TERMINATION
    assert engine.getFireRule("test.dat") == py_capio_cl.fire_rules.UPDATE
    assert engine.getConsumers("test.dat") == []
    assert engine.getProducers("test.dat") == []
    assert not engine.isPermanent("test.dat")
    assert not engine.isExcluded("test.dat")
    assert engine.isFile("test.dat")
    assert not engine.isDirectory("test.dat")
    assert engine.getDirectoryFileCount("test.dat") == 0
    assert not engine.isStoredInMemory("test.dat")
    engine.setFile("test.txt")


def test_add_file_manually_glob():
    engine = py_capio_cl.Engine()
    path = "test.*"
    engine.add(path, [], [], py_capio_cl.commit_rules.ON_TERMINATION,
               py_capio_cl.fire_rules.UPDATE, False, False, [])
    assert engine.size() == 1
    assert engine.getCommitRule("test.dat") == py_capio_cl.commit_rules.ON_TERMINATION
    assert engine.getFireRule("test.dat") == py_capio_cl.fire_rules.UPDATE
    assert engine.getConsumers("test.dat") == []
    assert engine.getProducers("test.dat") == []
    assert not engine.isPermanent("test.dat")
    assert not engine.isExcluded("test.dat")
    assert engine.isFile("test.dat")
    assert not engine.isDirectory("test.dat")
    assert engine.getDirectoryFileCount("test.dat") == 0
    assert not engine.isStoredInMemory("test.dat")


def test_add_file_manually_question():
    engine = py_capio_cl.Engine()
    path = "test.?"
    engine.add(path, [], [], py_capio_cl.commit_rules.ON_CLOSE,
               py_capio_cl.fire_rules.NO_UPDATE, False, False, [])
    engine.setDirectory("test.?")
    engine.setDirectoryFileCount("test.?", 10)

    assert engine.size() == 1
    assert engine.getCommitRule("test.1") == py_capio_cl.commit_rules.ON_CLOSE
    assert engine.getFireRule("test.1") == py_capio_cl.fire_rules.NO_UPDATE
    assert engine.isDirectory("test.1")
    assert engine.getDirectoryFileCount("test.?") == 10
    assert engine.getDirectoryFileCount("test.3") == 10
    assert engine.getConsumers("test.4") == []
    assert engine.getProducers("test.5") == []
    assert not engine.isPermanent("test.6")
    assert not engine.isExcluded("test.7")
    assert not engine.isFile("test.8")
    assert engine.isDirectory("test.9")
    assert engine.getDirectoryFileCount("test.a") == 10
    assert not engine.isStoredInMemory("test.b")

    engine.setDirectoryFileCount("myDir", 10)
    assert engine.getDirectoryFileCount("myDir") == 10


def test_add_file_manually_glob_explicit():
    engine = py_capio_cl.Engine()
    path = "test.[abc][abc][abc]"
    engine.add(path, [], [], py_capio_cl.commit_rules.ON_CLOSE,
               py_capio_cl.fire_rules.NO_UPDATE, False, False, [])
    engine.setDirectory(path)
    engine.setDirectoryFileCount(path, 10)
    assert engine.size() == 1
    assert engine.getCommitRule("test.cab") == py_capio_cl.commit_rules.ON_CLOSE
    assert engine.getFireRule("test.bac") == py_capio_cl.fire_rules.NO_UPDATE
    assert engine.isDirectory("test.bbb")


def test_producers_consumers_file_dependencies():
    engine = py_capio_cl.Engine()
    producers = ["A", "B"]
    consumers = ["C", "D"]
    deps = [PosixPath("E"), PosixPath("F")]

    engine.newFile("test.dat")
    engine.addProducer("test.dat", producers[0])
    assert engine.isProducer("test.dat", producers[0])
    engine.addProducer("test.dat", producers[1])
    assert engine.isProducer("test.dat", producers[1])

    engine.addConsumer("test.dat", consumers[0])
    assert engine.isConsumer("test.dat", consumers[0])
    engine.addConsumer("test.dat", consumers[1])
    assert engine.isConsumer("test.dat", consumers[1])

    assert engine.getCommitOnFileDependencies("test.dat") == []
    engine.addFileDependency("test.dat", deps[0])
    assert engine.getCommitOnFileDependencies("test.dat")[0] == deps[0]
    engine.addFileDependency("test.dat", deps[1])
    assert engine.getCommitOnFileDependencies("test.dat")[1] == deps[1]

    engine.addFileDependency("myFile.txt", deps[0])
    assert engine.getCommitRule("myFile.txt") == py_capio_cl.commit_rules.ON_FILE
    assert engine.getCommitOnFileDependencies("myFile.txt")[0] == deps[0]


def test_commit_fire_permanent_exclude_on_globs():
    engine = py_capio_cl.Engine()
    engine.newFile("test.*")
    engine.setFireRule("test.*", py_capio_cl.fire_rules.NO_UPDATE)
    assert engine.isFirable("test.a")
    assert not engine.isFirable("testb")

    engine.setCommitRule("testb", py_capio_cl.commit_rules.ON_FILE)
    engine.setFileDeps("testb", ["test.a"])

    engine.setPermanent("myFile", True)
    assert engine.isPermanent("myFile")
    assert not engine.isFirable("myFile")

    assert not engine.isExcluded("testb")
    engine.setExclude("testb", True)
    assert engine.isExcluded("testb")
    engine.setExclude("myFile.*", True)
    assert engine.isExcluded("myFile.txt")
    assert engine.isExcluded("myFile.dat")

    engine.setFireRule("test.c", py_capio_cl.fire_rules.NO_UPDATE)
    assert engine.isFirable("test.c")

    engine.setCommittedCloseNumber("test.e", 100)


def test_is_file_is_directory_glob():
    engine = py_capio_cl.Engine()
    engine.newFile("test.*")
    engine.setDirectory("test.d/")
    engine.setDirectory("test.d/bin/lib")
    assert engine.isDirectory("test.d/")
    assert engine.isDirectory("test.d/bin/lib")
    assert not engine.isDirectory("test.*")


def test_add_remove_file():
    engine = py_capio_cl.Engine()
    engine.newFile("test.*")
    assert engine.contains("test.*")
    assert engine.contains("test.txt")
    engine.remove("test.*")
    assert not engine.contains("test.*")
    engine.remove("data")
    assert not engine.contains("data")


def test_commit_close_count():
    engine = py_capio_cl.Engine()
    engine.newFile("test.*")
    engine.setCommitRule("test.*", py_capio_cl.commit_rules.ON_CLOSE)
    engine.setCommittedCloseNumber("test.e", 100)
    assert engine.getCommitCloseCount("test.e") == 100
    engine.setCommittedCloseNumber("test.*", 30)
    assert engine.getCommitCloseCount("test.f") == 30
    engine.setCommitRule("myFile", py_capio_cl.commit_rules.ON_FILE)
    assert engine.getCommitRule("myFile") == py_capio_cl.commit_rules.ON_FILE


def test_storage_options(tmp_path):
    engine = py_capio_cl.Engine()
    for f in ["A", "B"]:
        engine.newFile(f)

    engine.setStoreFileInMemory("A")
    assert engine.isStoredInMemory("A")
    assert not engine.isStoredInMemory("B")

    engine.setStoreFileInMemory("B")
    assert engine.isStoredInMemory("A")
    assert engine.isStoredInMemory("B")

    engine.setStoreFileInMemory("C")
    assert engine.isStoredInMemory("C")

    engine.newFile("D")
    assert not engine.isStoredInMemory("D")

    engine.setAllStoreInMemory()
    for f in ["A", "B", "C", "D"]:
        assert engine.isStoredInMemory(f)

    assert len(engine.getFileToStoreInMemory()) == 4

    for f in ["A", "B", "C", "D"]:
        engine.setStoreFileInFileSystem(f)
        assert not engine.isStoredInMemory(f)

    engine.setStoreFileInFileSystem("F")
    assert not engine.isStoredInMemory("F")


def test_home_node():
    """
    This test is skipped on macOS runners due to the issue described here:
    https://github.com/actions/runner-images/issues/10924.
    The test will only be executed on Linux-based runners, not macOS.
    """
    this_node_name = socket.gethostname()
    engine = py_capio_cl.Engine()
    engine.newFile("./A")
    assert this_node_name not in engine.getHomeNode("B")
    assert this_node_name not in engine.getHomeNode("./A")
    engine1 = py_capio_cl.Engine()
    engine1.setHomeNode("./A")
    time.sleep(1)
    assert this_node_name in engine.getHomeNode("./A")


def test_commit_status():
    """
    This test is skipped on macOS runners due to the issue described here:
    https://github.com/actions/runner-images/issues/10924.
    The test will only be executed on Linux-based runners, not macOS.
    """
    engine = py_capio_cl.Engine()
    engine.setCommitted("./C")
    assert engine.isCommitted("./C")

    engine1 = py_capio_cl.Engine()
    assert engine1.isCommitted("./C")


def test_insert_file_dependencies():
    engine = py_capio_cl.Engine()
    engine.setFileDeps("myFile.txt", [])
    engine.setFileDeps("test.txt", ["a", "b", "c"])
    deps = engine.getCommitOnFileDependencies("test.txt")
    assert deps == [PosixPath("a"), PosixPath("b"), PosixPath("c")]
